import os
import uuid
import asyncio
import unittest

from asyncrpc.tests import Test
from asyncrpc.server import UniCastServer
from asyncrpc.client import UniCastClient, RPCMethodException


class TestUniCastServer(unittest.TestCase):

    loop = asyncio.get_event_loop()

    def setUp(self):
        port = 9000
        self.interfaces_info = [('127.0.0.1', port), ('127.0.0.2', port)]
        ip_addrs = [ip_addr for ip_addr, _ in self.interfaces_info]
        self.srvc = UniCastServer(
            obj=Test(),
            ip_addrs=ip_addrs,
            port=port
        )
        self.srvc.delay = 0.1
        self.loop.run_until_complete(self.srvc.start())

        self.client = UniCastClient(interfaces_info=self.interfaces_info)

    def tearDown(self):
        self.loop.run_until_complete(self.client.close())
        self.loop.run_until_complete(self.srvc.stop())

    def test_rpc(self):
        msg = str(uuid.uuid1())
        result = self.loop.run_until_complete(self.client.echo(msg=msg))
        self.assertEquals(result, msg)

    def test_only_one_called(self):
        self.loop.run_until_complete(self.client.add())
        count = self.loop.run_until_complete(self.client.get_count())
        self.assertEquals(count, 1)

    def test_exception(self):
        with self.assertRaises(RPCMethodException):
            self.loop.run_until_complete(self.client.error())

    def test_func(self):
        msg = str(uuid.uuid1())
        result = self.loop.run_until_complete(self.client.func(msg))
        self.assertEquals(result, msg)


class TestUniCastServerWithSSLAuth(TestUniCastServer):

    def setUp(self):
        ssl_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'ssl')
        port = 9000
        self.interfaces_info = [('127.0.0.1', port), ('127.0.0.2', port)]
        ip_addrs = [ip_addr for ip_addr, _ in self.interfaces_info]
        self.srvc = UniCastServer(
            obj=Test(),
            ip_addrs=ip_addrs,
            port=port,
            cafile=os.path.join(ssl_path, 'ca.crt'),
            certfile=os.path.join(ssl_path, 'crt', 'server.crt'),
            keyfile=os.path.join(ssl_path, 'key', 'server.key')
        )
        self.srvc.delay = 0.1
        self.loop.run_until_complete(self.srvc.start())

        self.client = UniCastClient(
            interfaces_info=self.interfaces_info,
            certfile=os.path.join(ssl_path, 'crt', 'client01.crt'),
            keyfile=os.path.join(ssl_path, 'key', 'client01.key')
        )



import json
import uuid
import asyncio


class Test:

    count = 0

    @asyncio.coroutine
    def echo(self, msg):
        return msg

    @asyncio.coroutine
    def error(self):
        raise Exception('Internal Error')

    @asyncio.coroutine
    def add(self):
        self.count += 1

    @asyncio.coroutine
    def get_count(self):
        return self.count

    def func(self, msg):
        return msg


def create_request(method, *args, **kwargs):
    class Request:

        @asyncio.coroutine
        def read(self):
            return json.dumps(
                dict(
                    jsonrpc='2.0',
                    method=method,
                    params=args or kwargs,
                    id=str(uuid.uuid1())
                )
            ).encode()

    return Request()



import json
import uuid
import inspect
import aiohttp
import asyncio
import traceback
from collections import OrderedDict

from asyncrpc.cleaner import Cleaner


def _create_request(method, *args, **kwargs):
    return dict(
        jsonrpc='2.0',
        method=method,
        params=args or kwargs,
        id=str(uuid.uuid1())
    )


def create_request(method, *args, **kwargs):
    return serialize(_create_request(method, *args, **kwargs))


class RequestsStorage(Cleaner):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.events = OrderedDict()
        self.lock = asyncio.Lock()

    def is_called(self, request_id):
        return request_id in self.data

    def set_result(self, request_id):
        if request_id not in self.data:
            self.events[request_id] = asyncio.Event()
            self.data[request_id] = dict()

    @asyncio.coroutine
    def wait(self, request_id):
        yield from self.events[request_id].wait()

    def set(self, request_id):
        self.events[request_id].set()

    def popleft(self, count):
        for _ in range(count):
            self.events.popitem(last=False)

        super().popleft(count)


def serialize(value):
    return json.dumps(value).encode()


def deserialize(data):
    return json.loads(data.decode())


def isinteger(value):
    try:
        int(value)
        return True
    except ValueError:
        return False


@asyncio.coroutine
def _call_method(obj, storage, methods, request):
    try:
        with (yield from storage.lock):
            flag_called = storage.is_called(request['id'])
            storage.set_result(request['id'])

        if not isinteger(request['id']) and flag_called:
            yield from storage.wait(request['id'])
            respond = dict(
                jsonrpc='2.0',
                error='Method {} was already called'.format(request['method']),
                id=request['id']
            )
            return respond, 202

        if request['method'] not in methods:
            respond = dict(
                jsonrpc='2.0',
                error='Method {} does not exist'.format(request['method']),
                id=request['id']
            )
            storage.set(request['id'])
            return respond, 405

        method = getattr(obj, request['method'])

        if inspect.isgeneratorfunction(methods[request['method']]):
            if isinstance(request['params'], dict):
                result = yield from method(**request['params'])
            else:
                result = yield from method(*request['params'])
        else:
            if isinstance(request['params'], dict):
                result = method(**request['params'])
            else:
                result = method(*request['params'])

        respond = dict(
            jsonrpc='2.0',
            result=result,
            id=request['id']
        )
        storage.set(request['id'])
        return respond, 200
    except:
        respond = dict(
            jsonrpc='2.0',
            error=traceback.format_exc(),
            id=request['id']
        )
        storage.set(request['id'])
        return respond, 500

@asyncio.coroutine
def call_method(obj, storage, methods, request):
    storage.try_clear()
    data = yield from request.read()
    response, status = yield from _call_method(obj, storage, methods, deserialize(data))
    return aiohttp.web.Response(body=serialize(response), status=status)



import sys
import ssl
import time
import asyncio
from collections import deque

import aiohttp
from aiohttp.errors import ClientOSError

from asyncrpc.call import create_request, deserialize


PY_35 = sys.version_info >= (3, 5)


class _ContextManager:

    def __init__(self, client):
        self.client = client

    def __enter__(self):
        return None

    @asyncio.coroutine
    def __exit__(self, *args):
        try:
            yield from self.client.stop()
        finally:
            self.client = None  # Crudely prevent reuse.


class ContextManagerMixin:

    def __enter__(self):
        raise RuntimeError(
            '"yield from" should be used as context manager expression')

    def __exit__(self, *args):
        pass

    @asyncio.coroutine
    def __iter__(self):
        return _ContextManager(self)

    if PY_35:
        @asyncio.coroutine
        def __aenter__(self):
            return self

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_val, exc_tb):
            yield from self.close()


class LockCounter:
    _count = 0

    def __init__(self):
        self._fut = asyncio.Future()
        self._fut.set_result(None)

    def __enter__(self):
        self._count += 1
        self._fut = asyncio.Future()

    def __exit__(self, a, b, c):
        self._count -= 1
        if self._count <= 0:
            self._fut.set_result(None)

    if PY_35:
        @asyncio.coroutine
        def __aenter__(self):
            return self.__enter__()

        @asyncio.coroutine
        def __aexit__(self, *a):
            return self.__exit__(*a)

    @asyncio.coroutine
    def wait(self):
        yield from self._fut


class RequestError(Exception):
    pass


class RPCMethodException(Exception):

    def __init__(self, value):
        super().__init__(value)
        self.value = value


@asyncio.coroutine
def wait_first_completed(futures):
    flag_completed = False

    response = tuple()

    timeout = 30
    while not flag_completed:
        done, pending = (yield from asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED, timeout=timeout))

        if not done:
            for f in pending:
                f.cancel()
            raise RequestError()

        error_info = None
        for f in done:
            try:
                result, status, headers = f.result()
                response = result, status, headers
                if status != 200:
                    if status == 500:
                        error_info = result
                    continue
                flag_completed = True
                break
            except:
                pass

        timeout = 15

        if flag_completed:
            break

        if pending:
            futures = list(pending)
        else:
            try:
                if error_info is not None:
                    raise RPCMethodException(error_info)
                done.pop().result()  # raise exception
                return response + (pending,)
            except ClientOSError:
                raise RequestError()

    return result, status, headers, pending


def session_decorator(func):
    @asyncio.coroutine
    def session_post(self, data=None, headers=None, patch=None):
        futures = list()

        yield from self.try_clear()

        if not self.sessions:
            connector = aiohttp.TCPConnector(
                loop=self.loop, ssl_context=self.ssl_context, limit=self.limit
            )
            self.sessions.append(aiohttp.ClientSession(connector=connector))

        for ip_addr, port in self.interfaces_info:
            futures.append(
                func(
                    self,
                    url='{}/{}'.format(self.url_mask.format(ip_addr, port), patch or self.patch),
                    session=self.sessions[-1],
                    data=data,
                    headers=headers
                )
            )
        try:
            result, status, headers, pending = yield from wait_first_completed(futures)
        except RPCMethodException as error_data:
            for session in self.sessions:
                yield from session.close()
            self.sessions.clear()
            error = deserialize(error_data.value)
            raise RPCMethodException(error['error'])
        except RequestError:
            for session in self.sessions:
                yield from session.close()
            self.sessions.clear()
            raise RPCMethodException('RequestError')

        for f in pending:
            f.cancel()

        return result, status, headers
    return session_post


class UniCastClient(ContextManagerMixin):
    _closing = False

    def __init__(self, interfaces_info, patch='post', limit=20, loop=None, url_mask=None,
                 certfile=None, keyfile=None, **kwargs):
        self.interfaces_info = interfaces_info
        self.patch = patch

        self.limit = limit

        self.loop = asyncio.get_event_loop() if loop is None else loop

        self.url_mask = url_mask or 'http://{}:{}'

        self.sessions = deque()

        self.__start_ts = None
        self.__finish_ts = None
        self.monitoring_timeout = 5

        self._req_counter = LockCounter()

        self.ssl_context = None
        if certfile and keyfile:
            self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            self.ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)
            self.url_mask = url_mask or 'https://{}:{}'

    def __getattr__(self, name):
        return lambda *args, **kwargs: self._async_call(name, *args, **kwargs)

    @asyncio.coroutine
    def _async_call(self, method_name, *args, **kwargs):
        request = create_request(method_name, *args, **kwargs)
        data, _, _ = yield from self.session_post(data=request)
        response = deserialize(data)
        if 'error' in response:
            raise RPCMethodException(response['error'])
        return response['result']

    @asyncio.coroutine
    def request(self, f):
        if self._closing:
            raise asyncio.CancelledError

        with self._req_counter:
            resp = yield from f
            try:
                return (yield from resp.read()), resp.status, resp.headers
            finally:
                yield from resp.release()

    @session_decorator
    def session_post(self, url, session, data, headers):
        return (
            yield from self.request(session.post(url=url, data=data, headers=headers))
        )

    @session_decorator
    def session_delete(self, url, session, *args, **kwargs):
        return (
            yield from self.request(session.delete(url=url))
        )

    @session_decorator
    def session_get(self, url, session, headers, *args, **kwargs):
        return (
            yield from self.request(session.get(url=url, headers=headers))
        )

    @asyncio.coroutine
    def close(self):
        self._closing = True
        yield from self._req_counter.wait()
        for session in self.sessions:
            yield from session.close()
        self.sessions.clear()

    @asyncio.coroutine
    def try_clear(self):
        if self.__start_ts is None:
            self.__start_ts = time.perf_counter()
            return

        if time.perf_counter() - self.__start_ts > self.monitoring_timeout and self.__finish_ts is None:
            self.__finish_ts = time.perf_counter()

            connector = aiohttp.TCPConnector(loop=self.loop, ssl_context=self.ssl_context, limit=self.limit)
            self.sessions.append(aiohttp.ClientSession(connector=connector))
            return

        if self.__finish_ts and time.perf_counter() - self.__finish_ts > self.monitoring_timeout:
            if len(self.sessions) > 1:
                session = self.sessions.popleft()
                yield from session.close()
            self.__start_ts = None
            self.__finish_ts = None
            return



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name='asyncrpc',
    version=__import__('asyncrpc').__version__,
    description='JSON RPC Server and Client',
    author='Seliverstov Maksim',
    author_email='Maksim.V.Seliverstov@yandex.ru',
    packages=find_packages(),
    zip_safe=False,
    keywords=['rpc', 'jsonrpc', 'aiorpc', 'asyncrpc', 'multiple interfaces rpc'],
    install_requires=['aiohttp']
)



import uuid
import asyncio
import unittest
from unittest.mock import patch

from asyncrpc.tests import Test, create_request
from asyncrpc.call import call_method, RequestsStorage, deserialize


class Response:

    def __init__(self, body, *args, **kwargs):
        self.body = deserialize(body)


class TestJSONRPCCallMethod(unittest.TestCase):
    loop = asyncio.get_event_loop()

    @patch('aiohttp.web.Response', side_effect=Response)
    def test_call_method(self, _):
        msg = str(uuid.uuid1())
        request = create_request('echo', msg=msg)

        respond = self.loop.run_until_complete(
            call_method(Test(), RequestsStorage(), dict(echo=Test.echo), request)
        )

        self.assertIn('jsonrpc', respond.body)
        self.assertIn('result', respond.body)
        self.assertNotIn('error', respond.body)
        self.assertIn('id', respond.body)
        self.assertEquals(msg, respond.body['result'])

    @patch('aiohttp.web.Response', side_effect=Response)
    def test_error_call_method(self, _):
        request = create_request('error')

        respond = self.loop.run_until_complete(
            call_method(Test(), RequestsStorage(), dict(error=Test.error), request)
        )

        self.assertIn('jsonrpc', respond.body)
        self.assertNotIn('result', respond.body)
        self.assertIn('error', respond.body)
        self.assertIn('id', respond.body)

    @patch('aiohttp.web.Response', side_effect=Response)
    def test_does_not_exist_method(self, _):
        request = create_request('not_exist')

        respond = self.loop.run_until_complete(
            call_method(Test(), RequestsStorage(), dict(), request)
        )

        self.assertIn('jsonrpc', respond.body)
        self.assertNotIn('result', respond.body)
        self.assertIn('error', respond.body)
        self.assertIn('id', respond.body)
